import requests
import os
import json


repourl = "http://vps-407d02be.vps.ovh.us:5001"


def run_command(args):
	if not os.path.exists('kpm.json'):
		print("no kpm.json file found")
		return

	with open('kpm.json') as f:
		kpmjson = json.loads(f.read())

	kpmrc_path = '.kpmrc'
	if not os.path.exists(kpmrc_path):
		kpmrc_path = os.path.expanduser('~/.kpmrc')
	if not os.path.exists(kpmrc_path):
		print("no .kpmrc directory found or ~/.kpmrc directory found")

	with open(kpmrc_path) as f:
		kpmrc = json.loads(f.read())

	if 'token' not in kpmrc:
		print("kpmrc does not have a 'token' field")
		return

	if 'name' not in kpmrc:
		print("kpmrc does not have a 'name' field")
		return

	pkgname = kpmjson['name']
	version = kpmjson['version']

	r = requests.put(f"{repourl}/package/{pkgname}/release/{version}", files={'file': open('package.zip', 'rb')})
	if r.status_code != 200:
		print(f"failed to upload file: {r.content}")
		return
	artifacturl = r.content.decode('utf-8').strip()


	r = requests.get(f"{repourl}/package/{pkgname}")
	if r.status_code == 404:
		pkgdata = {
			"name": pkgname,
			"owner": kpmjson['author'], 
			"homepage": kpmjson['homepage'],
			"releases": []
		}
	elif r.status_code == 200:
		pkgdata = r.json()
	else:
		print(f"server error {r.status_code}: {r.content}")


	new_release = {
		"version": version, 
		"artifact_url": artifacturl, 
		"author": kpmjson['author'],
		"dependencies": kpmjson['dependencies']
	}
	pkgdata['releases'].append(new_release)
	print(json.dumps(pkgdata, indent=4))
	r = requests.put(f"{repourl}/package/{pkgname}", json=pkgdata, headers={"Authorization": kpmrc['token']})
	print('status', r.status_code)
	print('content', r.content)




import json
from . import config
import requests
from .registry import get_release_for, repourl, get as get_package
import os
import io
import zipfile
from . import kicad_project_tables
import glob
import shutil
from .init import init_kpmjson


def init_command(parser):
	parser.add_argument('package_ref')
	parser.add_argument('--version', '-v', type=str, required=False)

def run_command(args):
	package_ref = args.package_ref

	if package_ref != '.':
		if not os.path.exists('kpm.json'):
			init_kpmjson()

		with open('kpm.json') as kpmf:
			kpmjson = json.loads(kpmf.read())

		name = args.package_ref
		package = get_package(name)
		if package is None:
			print(f"Could not find a package by name {name}")
			return

		if args.version is None:
			version = package['releases'][-1]['version']
		else:
			version = args.version
			version_found = False
			for release in package['releases']:
				if release['version'] == version:
					version_found = True
			if not version_found:
				print(f"Could not find version {version} for package {name}. Did you mean {package['releases'][-1]['version']}?")
				return

		if name in kpmjson['dependencies']:
			if kpmjson['dependencies'][name] == version:
				print(f"Package {name} is already installed! run `kpm install .` to load all of your libraries")
				return

		kpmjson['dependencies'][name] = version

		with open('kpm.json', 'w') as kpmf:
			kpmf.write(json.dumps(kpmjson, indent=4))

	install_from_config()


def install_from_config():
		with open('kpm.json') as kpmf:
			kpmjson = json.loads(kpmf.read())
		deps = {}
		if 'dependencies' in kpmjson:
			for depname, depversion in kpmjson['dependencies'].items():
				explore_deps(depname, depversion, deps)
			install_deps(deps)


def explore_deps(name, version, found_packages={}):
	if name in found_packages:
		if found_packages[name] != version:
			raise Exception(f"incompatible version for {name}: {version}, committed version is {found_packages[name]['version']}")
		else:
			return found_packages

	release = get_release_for(name, version)

	if release is None:
		print(f"ERROR: version {version} not found for package {name}\n\n")
		raise Exception("package version missing")

	found_packages[name] = release

	if 'dependencies' in release:
		for depname, depversion in release['dependencies'].items():
			explore_deps(depname, depversion, found_packages)

	return found_packages


def install_deps(deps):
	shutil.rmtree("./kpm_modules", ignore_errors=True)
	for package_name, release in deps.items():
		print(f"installing {package_name} @ {release['version']}")
		install_package(package_name, release['version'], release['artifact_url'])
	install_libraries()


def install_package(name, version, zip_url):
	package_dir = f"./kpm_modules/{name}@{version}/"
	os.makedirs(package_dir, exist_ok=True)
	if zip_url[0] == '/':
		urlzip_url = repourl + zip_url
	res = requests.get(zip_url)
	r = zipfile.ZipFile(io.BytesIO(res.content)).extractall(package_dir)


def install_libraries():
	# link symbol files
	symfiles = glob.glob("kpm_modules/**/symbols/*.kicad_sym", recursive=True)
	kicad_project_tables.write_sym_lib_table(symfiles)

	# link footprint files
	footfiles = glob.glob("kpm_modules/**/footprints/*.pretty", recursive=True)
	kicad_project_tables.write_fp_lib_table(footfiles)

	# install 3d models
	# install spice models
	# install plugins




from . import registry


def init_command(parser):
	parser.add_argument('package_ref')


def run_command(args):
	packages = registry.search(args.package_ref)
	print_packages(packages)


def print_packages(packages):
	print(f"Matching packages from registry:")
	print()
	for package in packages:
		release = package['releases'][-1]
		print(f"{package['name']}\tv{release['version']} - {release['author']}")
	print()



from functools import lru_cache
import pathlib
import os
import requests


repourl = "http://vps-407d02be.vps.ovh.us:5001"


def get_release_for(package_name, version):
	package = get(package_name)
	if package is None:
		raise Exception(f"no package exists by name {package_name}")

	if version == 'latest':
		return package['releases'][-1]

	for release in package['releases']:
		if release['version'] == version:
			return release

	return None


def get_zip_url_for(package_name, version):
	release = get_release_for(package_name, version)
	url = release['artifact_url']
	if url[0] == '/':
		url = repourl + url
	return url


def search(package_name):
	r = requests.get(f"{repourl}/packages?term={package_name}")
	if r.status_code != 200:
		print("couldnt get packages list from registry")
		print(r)
		print(r.status_code)
		print(r.content)
		raise Exception("nooo")
	return r.json()


def get(package_name):
	r = requests.get(f"{repourl}/package/{package_name}")
	if r.status_code != 200:
		print("couldnt get package from registry")
		print(r)
		print(r.status_code)
		print(r.content)
		raise Exception("nooo")
	return r.json()


def listt():
	return search("")



from . import registry


def init_command(parser):
	parser.add_argument('package_ref')


def run_command(args):
	package = registry.get(args.package_ref)
	if package is None:
		print(f"No package found in our registry for '{args.package_ref}'")
		return
	print_package(args.package_ref, package)


def print_package(name, package):
	print(f"Package: {name}")
	print(f"Owner: {package['owner']}")
	print(f"Homepage: {package['homepage']}")
	print()
	print("Releases:")
	for release in package['releases']:
		print(f"v{release['version']} - {release['author']} - {release['artifact_url']}")
		print("  dependencies:")
		if 'dependencies' in release:
			for dep, version in release['dependencies'].items():
				print(f"  - {dep}@{version}")
	print()



import argparse
from . import install
from . import init
from . import listt
from . import search
from . import show
from . import build
from . import upload
from . import run
from . import init_package


def main():
	parser = argparse.ArgumentParser(
		prog="kpm",
		description="KiCad Package Manager",
		epilog="manage kicad parts and plugins good"
	)

	subparsers = parser.add_subparsers(dest='_command')
	
	commands = {
		"list":    listt,
		"search":  search,
		"show":    show,
		"init":    init,
		"install": install,
		"build":   build,
		"upload":  upload,
		"run":     run,
		"init_package": init_package,
	}
	for command_name, module in commands.items():
		subparser = subparsers.add_parser(command_name)
		if hasattr(module, 'init_command'):
			module.init_command(subparser)

	args = parser.parse_args()

	if args._command is None:
		parser.print_help()
		return

	for command_name, module in commands.items():
		if args._command == command_name:
			module.run_command(args)




import json
import os
from . import init

def init_command(parser):
	parser.add_argument('project_name')


def run_command(args):
	# make directory for project
	# make directory in that for each of the things
	# make a .gitkeep file in each directory
	# init with the kpm init function
	name = args.project_name
	os.makedirs(name)
	os.chdir(name)
	dirs = [
		'commands',
		'symbols',
		'footprints',
		'3dmodels',
		'plugins',
		'simulation',
		'subsheets',
		'tests',
		'scripts',
	]
	for d in dirs:
		os.makedirs(d)
		with open(os.path.join(d, ".gitkeep"), 'w') as f:
			f.write("keep")

	init.init_kpmjson()

	# also make example git workflows

	os.chdir('..')



import os
import shutil
import zipfile


def init_command(parser):
    parser.add_argument('--keep-build-dir', '-k', action="store_true", required=False)


def zipdir(path, ziph):
    for root, dirs, files in os.walk(path):
        for file in files:
            from_path = os.path.join(root, file)
            to_path = os.path.relpath(os.path.join(root, file), os.path.join(path, '..')).replace(path + "/","",1)
            print("adding", from_path)
            ziph.write(from_path, to_path)


def run_command(args):
    build_dir = 'build'
    if os.path.exists(build_dir):
        shutil.rmtree(build_dir)
    os.makedirs(build_dir)

    package_dirs = ['symbols', 'footprints', 'subsheets', 'commands']
    for package_dir in package_dirs:
        if os.path.exists(package_dir):
            shutil.copytree(package_dir, os.path.join(build_dir, package_dir))
        else:
            print(f"no {package_dir}/ directory, skipping")
    shutil.copy('kpm.json', build_dir)

    with zipfile.ZipFile('package.zip', 'w', zipfile.ZIP_LZMA) as zh:
        zipdir(build_dir, zh)

    if not args.keep_build_dir:
        shutil.rmtree(build_dir)




import glob
from importlib.machinery import SourceFileLoader
import os


commands = []
parser = None


def init_command(_parser):
	global parser
	parser = _parser
	subparsers = parser.add_subparsers(dest="command")

	init_files = glob.glob('kpm_modules/*/commands/__init__.py')
	for d in init_files:
		modulename = os.path.basename(os.path.dirname(d)).split("@")[0].replace("-","_")
		module = SourceFileLoader(modulename, d).load_module()

		if hasattr(module, 'register'):
			commands.append(module)
			module.register(subparsers)


def run_command(args):
	if args.command is None:
		parser.print_help()
		return

	for command_name, module in commands.items():
		if hasattr(module, 'run_command'):
			module.run_command(args)




import re
import glob
import json
import os
from . import registry
from . import config


def run_command(args):
    init_kpmjson()


def init_kpmjson(search_existing=True):
    modules = set()

    if search_existing:
        # get footprints from schematics
        for fname in glob.glob('**.kicad_sch'):
            with open(fname) as f:
                text = f.read()
                footprints = re.findall(r'\(\s*property\s+"Footprint"\s+"[^"]+"', text)
                footprints = [ re.sub('^.* "Footprint" "', '', fp).replace('"','') for fp in footprints]
                for ft in footprints:
                    modules.add("kicad_lib_footprints_" + re.sub(':.*', '', ft))

            symbols = re.findall(r'\(\s*symbol\s+"[^"]+:', text)
            symbols = [ re.sub('^.*"', '', s).replace(':','') for s in symbols]
            for s in symbols:
                modules.add("kicad_lib_symbols_" + s)

        # get footprints from pcb files
        for fname in glob.glob('**.kicad_pcb'):
            with open(fname) as f:
                text = f.read()
                footprints = re.findall(r'\(\s*footprint\s+"[^"]+"', text)
                footprints = [ re.sub('.*"', '', re.sub(':.*', '', fp)) for fp in footprints]
                for ft in footprints:
                    modules.add("kicad_lib_footprints_" + ft)

    deps = {}
    for modulename in sorted(modules):
        package = registry.get(modulename)
        if package:
            deps[modulename] = package['releases'][-1]['version']

    project_name = os.path.basename(os.getcwd())

    c = config.Config({
        "name": project_name,
        "dependencies": deps
    })

    # create our kpm.json config file
    with open('kpm.json', 'w') as f:
        f.write(c.toJSON())

    # add to or create .gitignore
    if os.path.exists('.gitignore'):
        with open('.gitignore') as fr:
            if 'kpm_modules' not in fr.read():
                with open('.gitignore', 'a') as fw:
                    fw.write("\nkpm_modules")
                    fw.write("\n.kpmrc")
                    fw.write("\npackage.zip")
                    fw.write("\nbuild")
    else:
        with open('.gitignore', 'w') as f:
            f.write("kpm_modules\n")



